"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _Parser_instances, _Parser_projectionStop, _Parser_cache, _Parser_maxCacheSize, _Parser_tokenizer, _Parser_tokens, _Parser_index, _Parser_doParse, _Parser_parse, _Parser_expression, _Parser_getNudFunction, _Parser_processQuotedIdentifier, _Parser_processStarToken, _Parser_processLParenTokenNud, _Parser_processFlattenTokenNud, _Parser_processLBracketTokenNud, _Parser_processDefaultToken, _Parser_getLedFunction, _Parser_processDotToken, _Parser_processPipeToken, _Parser_processOrToken, _Parser_processAndToken, _Parser_processLParenToken, _Parser_processFilterToken, _Parser_processFlattenToken, _Parser_processLBracketToken, _Parser_throwParseError, _Parser_parseIndexExpression, _Parser_parseSliceExpression, _Parser_projectIfSlice, _Parser_parseComparator, _Parser_parseMultiSelectList, _Parser_parseMultiSelectHash, _Parser_parseProjectionRhs, _Parser_parseDotRhs, _Parser_match, _Parser_matchMultipleTokens, _Parser_advance, _Parser_currentToken, _Parser_lookahead, _Parser_lookaheadToken, _Parser_evictCache;
Object.defineProperty(exports, "__esModule", { value: true });
exports.Parser = void 0;
const node_crypto_1 = require("node:crypto");
const Lexer_js_1 = require("./Lexer.js");
const ParsedResult_js_1 = require("./ParsedResult.js");
const ast_js_1 = require("./ast.js");
const constants_js_1 = require("./constants.js");
const errors_js_1 = require("./errors.js");
/**
 * Top down operaotr precedence parser for JMESPath.
 *
 * ## References
 * The implementation of this Parser is based on the implementation of
 * [jmespath.py](https://github.com/jmespath/jmespath.py/), which in turn
 * is based on [Vaughan R. Pratt's "Top Down Operator Precedence"](http://dl.acm.org/citation.cfm?doid=512927.512931).
 *
 * If you don't want to read the full paper, there are some other good
 * overviews that explain the general idea:
 * - [Pratt Parsers: Expression Parsing Made Easy](https://journal.stuffwithstuff.com/2011/03/19/pratt-parsers-expression-parsing-made-easy/)
 * - [Simple Top-Down Parsing in Python](https://11l-lang.org/archive/simple-top-down-parsing/)
 * - [Top Down Operator Precedence](http://javascript.crockford.com/tdop/tdop.html)
 */
class Parser {
    constructor(lookahead = 2) {
        _Parser_instances.add(this);
        /**
         * The maximum binding power for a token
         * that can stop a projection.
         */
        _Parser_projectionStop.set(this, 10);
        /**
         * Cache object
         */
        _Parser_cache.set(this, {});
        /**
         * The maximum size of the cache.
         */
        _Parser_maxCacheSize.set(this, 128);
        _Parser_tokenizer.set(this, void 0);
        _Parser_tokens.set(this, void 0);
        _Parser_index.set(this, 0);
        __classPrivateFieldSet(this, _Parser_tokens, Array.from({ length: lookahead }), "f");
    }
    /**
     * Parse a JMESPath expression and return the Abstract Syntax Tree (AST)
     * that represents the expression.
     *
     * The AST is cached, so if you parse the same expression multiple times,
     * the AST will be returned from the cache.
     *
     * @param expression The JMESPath expression to parse.
     */
    parse(expression) {
        const cached = __classPrivateFieldGet(this, _Parser_cache, "f")[expression];
        if (cached) {
            return cached;
        }
        const parsedResult = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_doParse).call(this, expression);
        __classPrivateFieldGet(this, _Parser_cache, "f")[expression] = parsedResult;
        if (Object.keys(__classPrivateFieldGet(this, _Parser_cache, "f")).length > __classPrivateFieldGet(this, _Parser_maxCacheSize, "f")) {
            __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_evictCache).call(this);
        }
        return parsedResult;
    }
    /**
     * Purge the entire cache.
     */
    purgeCache() {
        __classPrivateFieldSet(this, _Parser_cache, {}, "f");
    }
}
exports.Parser = Parser;
_Parser_projectionStop = new WeakMap(), _Parser_cache = new WeakMap(), _Parser_maxCacheSize = new WeakMap(), _Parser_tokenizer = new WeakMap(), _Parser_tokens = new WeakMap(), _Parser_index = new WeakMap(), _Parser_instances = new WeakSet(), _Parser_doParse = function _Parser_doParse(expression) {
    try {
        return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parse).call(this, expression);
    }
    catch (error) {
        if (error instanceof errors_js_1.LexerError ||
            error instanceof errors_js_1.IncompleteExpressionError ||
            error instanceof errors_js_1.ParseError) {
            error.setExpression(expression);
            throw error;
        }
        throw error;
    }
}, _Parser_parse = function _Parser_parse(expression) {
    __classPrivateFieldSet(this, _Parser_tokenizer, new Lexer_js_1.Lexer(), "f");
    __classPrivateFieldSet(this, _Parser_tokens, [...__classPrivateFieldGet(this, _Parser_tokenizer, "f").tokenize(expression)], "f");
    __classPrivateFieldSet(this, _Parser_index, 0, "f");
    const parsed = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, 0);
    if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) !== 'eof') {
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_throwParseError).call(this);
    }
    return new ParsedResult_js_1.ParsedResult(expression, parsed);
}, _Parser_expression = function _Parser_expression(bindingPower = 0) {
    const leftToken = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookaheadToken).call(this, 0);
    __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_advance).call(this);
    let left = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_getNudFunction).call(this, leftToken);
    let currentToken = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this);
    while (bindingPower < constants_js_1.BINDING_POWER[currentToken]) {
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_advance).call(this);
        left = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_getLedFunction).call(this, currentToken, left);
        currentToken = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this);
    }
    return left;
}, _Parser_getNudFunction = function _Parser_getNudFunction(token) {
    const { type: tokenType } = token;
    switch (tokenType) {
        case 'literal':
            return (0, ast_js_1.literal)(token.value);
        case 'unquoted_identifier':
            return (0, ast_js_1.field)(token.value);
        case 'quoted_identifier':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processQuotedIdentifier).call(this, token);
        case 'star':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processStarToken).call(this);
        case 'filter':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_getLedFunction).call(this, tokenType, (0, ast_js_1.identity)());
        case 'lbrace':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseMultiSelectHash).call(this);
        case 'lparen':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processLParenTokenNud).call(this);
        case 'flatten':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processFlattenTokenNud).call(this);
        case 'not':
            return (0, ast_js_1.notExpression)(__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, constants_js_1.BINDING_POWER['not']));
        case 'lbracket':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processLBracketTokenNud).call(this);
        case 'current':
            return (0, ast_js_1.currentNode)();
        case 'expref':
            return (0, ast_js_1.expref)(__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, constants_js_1.BINDING_POWER['expref']));
        default:
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processDefaultToken).call(this, token);
    }
}, _Parser_processQuotedIdentifier = function _Parser_processQuotedIdentifier(token) {
    const fieldValue = (0, ast_js_1.field)(token.value);
    if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) === 'lparen') {
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_throwParseError).call(this, {
            lexPosition: 0,
            reason: 'quoted identifiers cannot be used as a function name',
        });
    }
    return fieldValue;
}, _Parser_processStarToken = function _Parser_processStarToken() {
    const left = (0, ast_js_1.identity)();
    let right;
    if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) === 'rbracket') {
        right = (0, ast_js_1.identity)();
    }
    else {
        right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseProjectionRhs).call(this, constants_js_1.BINDING_POWER['star']);
    }
    return (0, ast_js_1.valueProjection)(left, right);
}, _Parser_processLParenTokenNud = function _Parser_processLParenTokenNud() {
    const expression = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this);
    __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'rparen');
    return expression;
}, _Parser_processFlattenTokenNud = function _Parser_processFlattenTokenNud() {
    const left = (0, ast_js_1.flatten)((0, ast_js_1.identity)());
    const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseProjectionRhs).call(this, constants_js_1.BINDING_POWER['flatten']);
    return (0, ast_js_1.projection)(left, right);
}, _Parser_processLBracketTokenNud = function _Parser_processLBracketTokenNud() {
    if (['number', 'colon'].includes(__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this))) {
        const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseIndexExpression).call(this);
        return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_projectIfSlice).call(this, (0, ast_js_1.identity)(), right);
    }
    else if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) === 'star' &&
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookahead).call(this, 1) === 'rbracket') {
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_advance).call(this);
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_advance).call(this);
        const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseProjectionRhs).call(this, constants_js_1.BINDING_POWER['star']);
        return (0, ast_js_1.projection)((0, ast_js_1.identity)(), right);
    }
    else {
        return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseMultiSelectList).call(this);
    }
}, _Parser_processDefaultToken = function _Parser_processDefaultToken(token) {
    if (token.type === 'eof') {
        throw new errors_js_1.IncompleteExpressionError({
            lexPosition: token.start,
            tokenValue: token.value,
            tokenType: token.type,
        });
    }
    throw new errors_js_1.ParseError({
        lexPosition: token.start,
        tokenValue: token.value,
        tokenType: token.type,
    });
}, _Parser_getLedFunction = function _Parser_getLedFunction(tokenType, leftNode) {
    switch (tokenType) {
        case 'dot':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processDotToken).call(this, leftNode);
        case 'pipe':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processPipeToken).call(this, leftNode);
        case 'or':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processOrToken).call(this, leftNode);
        case 'and':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processAndToken).call(this, leftNode);
        case 'lparen':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processLParenToken).call(this, leftNode);
        case 'filter':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processFilterToken).call(this, leftNode);
        case 'eq':
        case 'ne':
        case 'gt':
        case 'gte':
        case 'lt':
        case 'lte':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseComparator).call(this, leftNode, tokenType);
        case 'flatten':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processFlattenToken).call(this, leftNode);
        case 'lbracket':
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_processLBracketToken).call(this, leftNode);
        default:
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_throwParseError).call(this);
    }
}, _Parser_processDotToken = function _Parser_processDotToken(leftNode) {
    if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) !== 'star') {
        const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseDotRhs).call(this, constants_js_1.BINDING_POWER['dot']);
        if (leftNode.type === 'subexpression') {
            leftNode.children.push(right);
            return leftNode;
        }
        else {
            return (0, ast_js_1.subexpression)([leftNode, right]);
        }
    }
    else {
        // We are creating a value projection
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_advance).call(this);
        const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseProjectionRhs).call(this, constants_js_1.BINDING_POWER['dot']);
        return (0, ast_js_1.valueProjection)(leftNode, right);
    }
}, _Parser_processPipeToken = function _Parser_processPipeToken(leftNode) {
    const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, constants_js_1.BINDING_POWER['pipe']);
    return (0, ast_js_1.pipe)(leftNode, right);
}, _Parser_processOrToken = function _Parser_processOrToken(leftNode) {
    const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, constants_js_1.BINDING_POWER['or']);
    return (0, ast_js_1.orExpression)(leftNode, right);
}, _Parser_processAndToken = function _Parser_processAndToken(leftNode) {
    const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, constants_js_1.BINDING_POWER['and']);
    return (0, ast_js_1.andExpression)(leftNode, right);
}, _Parser_processLParenToken = function _Parser_processLParenToken(leftNode) {
    const name = leftNode.value;
    const args = [];
    while (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) !== 'rparen') {
        const expression = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this);
        if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) === 'comma') {
            __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'comma');
        }
        args.push(expression);
    }
    __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'rparen');
    return (0, ast_js_1.functionExpression)(name, args);
}, _Parser_processFilterToken = function _Parser_processFilterToken(leftNode) {
    // Filters are projections
    const condition = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, 0);
    __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'rbracket');
    let right;
    if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) === 'flatten') {
        right = (0, ast_js_1.identity)();
    }
    else {
        right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseProjectionRhs).call(this, constants_js_1.BINDING_POWER['flatten']);
    }
    return (0, ast_js_1.filterProjection)(leftNode, right, condition);
}, _Parser_processFlattenToken = function _Parser_processFlattenToken(leftNode) {
    const left = (0, ast_js_1.flatten)(leftNode);
    const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseProjectionRhs).call(this, constants_js_1.BINDING_POWER['flatten']);
    return (0, ast_js_1.projection)(left, right);
}, _Parser_processLBracketToken = function _Parser_processLBracketToken(leftNode) {
    const token = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookaheadToken).call(this, 0);
    if (['number', 'colon'].includes(token.type)) {
        const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseIndexExpression).call(this);
        if (leftNode.type === 'index_expression') {
            // Optimization: if the left node is an index expression
            // we can avoid creating another node and instead just
            // add the right node as a child of the left node.
            leftNode.children.push(right);
            return leftNode;
        }
        else {
            return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_projectIfSlice).call(this, leftNode, right);
        }
    }
    else {
        // We have a projection
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'star');
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'rbracket');
        const right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseProjectionRhs).call(this, constants_js_1.BINDING_POWER['star']);
        return (0, ast_js_1.projection)(leftNode, right);
    }
}, _Parser_throwParseError = function _Parser_throwParseError(options) {
    const token = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookaheadToken).call(this, 0);
    throw new errors_js_1.ParseError({
        lexPosition: options?.lexPosition ?? token.start,
        tokenValue: options?.tokenValue ?? token.value,
        tokenType: options?.tokenType ?? token.type,
        reason: options?.reason,
    });
}, _Parser_parseIndexExpression = function _Parser_parseIndexExpression() {
    // We're here:
    // [<current>
    //  ^
    //  | (currentToken)
    if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookahead).call(this, 0) === 'colon' || __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookahead).call(this, 1) === 'colon') {
        return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseSliceExpression).call(this);
    }
    else {
        // Parse the syntax [number]
        const node = (0, ast_js_1.index)(__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookaheadToken).call(this, 0).value);
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_advance).call(this);
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'rbracket');
        return node;
    }
}, _Parser_parseSliceExpression = function _Parser_parseSliceExpression() {
    // [start:end:step]
    // Where start, end, and step are optional.
    // The last colon is optional as well.
    const parts = [];
    let index = 0;
    let currentToken = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this);
    while (currentToken !== 'rbracket' && index < 3) {
        if (currentToken === 'colon') {
            index += 1;
            if (index === 3) {
                __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_throwParseError).call(this);
            }
            __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_advance).call(this);
        }
        else if (currentToken === 'number') {
            parts[index] = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookaheadToken).call(this, 0).value;
            __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_advance).call(this);
        }
        else {
            __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_throwParseError).call(this);
        }
        currentToken = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this);
    }
    __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'rbracket');
    return (0, ast_js_1.slice)(parts[0], parts[1], parts[2]);
}, _Parser_projectIfSlice = function _Parser_projectIfSlice(left, right) {
    const idxExpression = (0, ast_js_1.indexExpression)([left, right]);
    if (right.type === 'slice') {
        return (0, ast_js_1.projection)(idxExpression, __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseProjectionRhs).call(this, constants_js_1.BINDING_POWER['star']));
    }
    else {
        return idxExpression;
    }
}, _Parser_parseComparator = function _Parser_parseComparator(left, comparatorChar) {
    return (0, ast_js_1.comparator)(comparatorChar, left, __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, constants_js_1.BINDING_POWER[comparatorChar]));
}, _Parser_parseMultiSelectList = function _Parser_parseMultiSelectList() {
    const expressions = [];
    while (true) {
        const expression = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this);
        expressions.push(expression);
        if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) === 'rbracket') {
            break;
        }
        else {
            __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'comma');
        }
    }
    __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'rbracket');
    return (0, ast_js_1.multiSelectList)(expressions);
}, _Parser_parseMultiSelectHash = function _Parser_parseMultiSelectHash() {
    const pairs = [];
    while (true) {
        const keyToken = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookaheadToken).call(this, 0);
        // Before getting the token value, verify it's
        // an identifier.
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_matchMultipleTokens).call(this, ['quoted_identifier', 'unquoted_identifier']); // token types
        const keyName = keyToken['value'];
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'colon');
        const value = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, 0);
        const node = (0, ast_js_1.keyValPair)(keyName, value);
        pairs.push(node);
        if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) == 'comma') {
            __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'comma');
        }
        else if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) == 'rbrace') {
            __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'rbrace');
            break;
        }
    }
    return (0, ast_js_1.multiSelectObject)(pairs);
}, _Parser_parseProjectionRhs = function _Parser_parseProjectionRhs(bindingPower) {
    // Parse the right hand side of the projection.
    let right;
    if (constants_js_1.BINDING_POWER[__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this)] < __classPrivateFieldGet(this, _Parser_projectionStop, "f")) {
        // BP of 10 are all the tokens that stop a projection.
        right = (0, ast_js_1.identity)();
    }
    else if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) == 'lbracket') {
        right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, bindingPower);
    }
    else if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) == 'filter') {
        right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, bindingPower);
    }
    else if (__classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this) == 'dot') {
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'dot');
        right = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseDotRhs).call(this, bindingPower);
    }
    else {
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_throwParseError).call(this);
    }
    return right;
}, _Parser_parseDotRhs = function _Parser_parseDotRhs(bindingPower) {
    // From the grammar:
    // expression '.' ( identifier /
    //                  multi-select-list /
    //                  multi-select-hash /
    //                  function-expression /
    //                  *
    // In terms of tokens that means that after a '.',
    // you can have:
    const lookahead = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this);
    // Common case "foo.bar", so first check for an identifier.
    if (['quoted_identifier', 'unquoted_identifier', 'star'].includes(lookahead)) {
        return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_expression).call(this, bindingPower);
    }
    else if (lookahead == 'lbracket') {
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'lbracket');
        return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseMultiSelectList).call(this);
    }
    else if (lookahead == 'lbrace') {
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_match).call(this, 'lbrace');
        return __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_parseMultiSelectHash).call(this);
    }
    else {
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_throwParseError).call(this);
    }
}, _Parser_match = function _Parser_match(tokenType) {
    const currentToken = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this);
    if (currentToken === tokenType) {
        __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_advance).call(this);
    }
    else {
        const token = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookaheadToken).call(this, 0);
        if (token.type === 'eof') {
            throw new errors_js_1.IncompleteExpressionError({
                lexPosition: token.start,
                tokenValue: token.value,
                tokenType: token.type,
            });
        }
        else {
            throw new errors_js_1.ParseError({
                lexPosition: token.start,
                tokenValue: token.value,
                tokenType: token.type,
            });
        }
    }
}, _Parser_matchMultipleTokens = function _Parser_matchMultipleTokens(tokenTypes) {
    const currentToken = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_currentToken).call(this);
    if (!tokenTypes.includes(currentToken)) {
        const token = __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_lookaheadToken).call(this, 0);
        if (token.type === 'eof') {
            throw new errors_js_1.IncompleteExpressionError({
                lexPosition: token.start,
                tokenValue: token.value,
                tokenType: token.type,
            });
        }
        else {
            throw new errors_js_1.ParseError({
                lexPosition: token.start,
                tokenValue: token.value,
                tokenType: token.type,
            });
        }
    }
    __classPrivateFieldGet(this, _Parser_instances, "m", _Parser_advance).call(this);
}, _Parser_advance = function _Parser_advance() {
    __classPrivateFieldSet(this, _Parser_index, __classPrivateFieldGet(this, _Parser_index, "f") + 1, "f");
}, _Parser_currentToken = function _Parser_currentToken() {
    return __classPrivateFieldGet(this, _Parser_tokens, "f")[__classPrivateFieldGet(this, _Parser_index, "f")].type;
}, _Parser_lookahead = function _Parser_lookahead(number) {
    return __classPrivateFieldGet(this, _Parser_tokens, "f")[__classPrivateFieldGet(this, _Parser_index, "f") + number].type;
}, _Parser_lookaheadToken = function _Parser_lookaheadToken(number) {
    return __classPrivateFieldGet(this, _Parser_tokens, "f")[__classPrivateFieldGet(this, _Parser_index, "f") + number];
}, _Parser_evictCache = function _Parser_evictCache() {
    const newCache = Object.keys(__classPrivateFieldGet(this, _Parser_cache, "f")).reduce((acc, key) => {
        if ((0, node_crypto_1.randomInt)(0, 100) > 50) {
            acc[key] = __classPrivateFieldGet(this, _Parser_cache, "f")[key];
        }
        return acc;
    }, {});
    __classPrivateFieldSet(this, _Parser_cache, newCache, "f");
};
