"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _TreeInterpreter_instances, _TreeInterpreter_functions, _TreeInterpreter_visitSubexpressionOrIndexExpressionOrPipe, _TreeInterpreter_visitField, _TreeInterpreter_visitComparator, _TreeInterpreter_visitCurrent, _TreeInterpreter_visitExpref, _TreeInterpreter_visitFunctionExpression, _TreeInterpreter_visitFilterProjection, _TreeInterpreter_visitFlatten, _TreeInterpreter_visitIdentity, _TreeInterpreter_visitIndex, _TreeInterpreter_visitSlice, _TreeInterpreter_visitKeyValPair, _TreeInterpreter_visitLiteral, _TreeInterpreter_visitMultiSelectObject, _TreeInterpreter_visitMultiSelectList, _TreeInterpreter_visitOrExpression, _TreeInterpreter_visitAndExpression, _TreeInterpreter_visitNotExpression, _TreeInterpreter_visitProjection, _TreeInterpreter_visitValueProjection;
Object.defineProperty(exports, "__esModule", { value: true });
exports.TreeInterpreter = void 0;
const typeutils_1 = require("@aws-lambda-powertools/commons/typeutils");
const errors_js_1 = require("./errors.js");
const Expression_js_1 = require("./Expression.js");
const Functions_js_1 = require("./Functions.js");
const utils_js_1 = require("./utils.js");
/**
 *
 * A tree interpreter for JMESPath ASTs.
 *
 * The tree interpreter is responsible for visiting nodes in the AST and
 * evaluating them to produce a result.
 *
 * @internal
 */
class TreeInterpreter {
    /**
     * @param options The options to use for the interpreter.
     */
    constructor(options) {
        _TreeInterpreter_instances.add(this);
        _TreeInterpreter_functions.set(this, void 0);
        if (options?.customFunctions) {
            __classPrivateFieldSet(this, _TreeInterpreter_functions, options.customFunctions, "f");
        }
        else {
            __classPrivateFieldSet(this, _TreeInterpreter_functions, new Functions_js_1.Functions(), "f");
        }
        __classPrivateFieldGet(this, _TreeInterpreter_functions, "f").introspectMethods();
    }
    /**
     * Visit a node in the AST.
     *
     * The function will call the appropriate method to visit the node based on its type.
     *
     * @param node The node to visit.
     * @param value The current value to visit.
     */
    visit(node, value) {
        const nodeType = node.type;
        const visitMethods = {
            subexpression: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitSubexpressionOrIndexExpressionOrPipe),
            field: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitField),
            comparator: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitComparator),
            current: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitCurrent),
            expref: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitExpref),
            function_expression: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitFunctionExpression),
            filter_projection: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitFilterProjection),
            flatten: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitFlatten),
            identity: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitIdentity),
            index: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitIndex),
            index_expression: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitSubexpressionOrIndexExpressionOrPipe),
            slice: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitSlice),
            key_val_pair: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitKeyValPair),
            literal: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitLiteral),
            multi_select_object: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitMultiSelectObject),
            multi_select_list: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitMultiSelectList),
            or_expression: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitOrExpression),
            and_expression: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitAndExpression),
            not_expression: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitNotExpression),
            pipe: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitSubexpressionOrIndexExpressionOrPipe),
            projection: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitProjection),
            value_projection: __classPrivateFieldGet(this, _TreeInterpreter_instances, "m", _TreeInterpreter_visitValueProjection),
        };
        const visitMethod = visitMethods[nodeType];
        if (visitMethod) {
            return visitMethod.call(this, node, value);
        }
        else {
            throw new errors_js_1.JMESPathError(`Not Implemented: Invalid node type: ${node.type}`);
        }
    }
}
exports.TreeInterpreter = TreeInterpreter;
_TreeInterpreter_functions = new WeakMap(), _TreeInterpreter_instances = new WeakSet(), _TreeInterpreter_visitSubexpressionOrIndexExpressionOrPipe = function _TreeInterpreter_visitSubexpressionOrIndexExpressionOrPipe(node, value) {
    let result = value;
    for (const child of node.children) {
        result = this.visit(child, result);
    }
    return result;
}, _TreeInterpreter_visitField = function _TreeInterpreter_visitField(node, value) {
    if (!node.value)
        return null;
    if ((0, typeutils_1.isRecord)(value) &&
        typeof node.value === 'string' &&
        node.value in value) {
        return value[node.value];
    }
    else {
        return null;
    }
}, _TreeInterpreter_visitComparator = function _TreeInterpreter_visitComparator(node, value) {
    const comparator = node.value;
    const left = this.visit(node.children[0], value);
    const right = this.visit(node.children[1], value);
    if (typeof comparator === 'string' &&
        ['eq', 'ne', 'gt', 'gte', 'lt', 'lte'].includes(comparator)) {
        // Common cases: comparator is == or !=
        if (comparator === 'eq') {
            return (0, typeutils_1.isStrictEqual)(left, right);
        }
        else if (comparator === 'ne') {
            return !(0, typeutils_1.isStrictEqual)(left, right);
        }
        else if (typeof left === 'number' && typeof right === 'number') {
            // Ordering operators only work on numbers. Evaluating them on other
            // types will return null.
            if (comparator === 'lt') {
                return left < right;
            }
            else if (comparator === 'lte') {
                return left <= right;
            }
            else if (comparator === 'gt') {
                return left > right;
            }
            else {
                return left >= right;
            }
        }
    }
    else {
        throw new errors_js_1.JMESPathError(`Invalid comparator: ${comparator}`);
    }
}, _TreeInterpreter_visitCurrent = function _TreeInterpreter_visitCurrent(_node, value) {
    return value;
}, _TreeInterpreter_visitExpref = function _TreeInterpreter_visitExpref(node, _value) {
    return new Expression_js_1.Expression(node.children[0], this);
}, _TreeInterpreter_visitFunctionExpression = function _TreeInterpreter_visitFunctionExpression(node, value) {
    const args = [];
    for (const child of node.children) {
        args.push(this.visit(child, value));
    }
    // check that method name is a string
    if (typeof node.value !== 'string') {
        throw new errors_js_1.JMESPathError(`Function name must be a string, got ${node.value}`);
    }
    // convert snake_case to camelCase
    const normalizedFunctionName = node.value.replace(/_([a-z])/g, (g) => g[1].toUpperCase());
    // capitalize first letter & add `func` prefix
    const funcName = `func${normalizedFunctionName.charAt(0).toUpperCase() +
        normalizedFunctionName.slice(1)}`;
    if (!__classPrivateFieldGet(this, _TreeInterpreter_functions, "f").methods.has(funcName)) {
        throw new errors_js_1.UnknownFunctionError(node.value);
    }
    try {
        // We know that methodName is a key of this.#functions, but TypeScript
        // doesn't know that, so we have to use @ts-ignore to tell it that it's
        // okay. We could use a type assertion like `as keyof Functions`, but
        // we also want to keep the args generic, so for now we'll just ignore it.
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore-next-line
        return __classPrivateFieldGet(this, _TreeInterpreter_functions, "f")[funcName](args);
    }
    catch (error) {
        if (error instanceof errors_js_1.JMESPathTypeError ||
            error instanceof errors_js_1.VariadicArityError ||
            error instanceof errors_js_1.ArityError) {
            error.setEvaluatedFunctionName(node.value);
            throw error;
        }
    }
}, _TreeInterpreter_visitFilterProjection = function _TreeInterpreter_visitFilterProjection(node, value) {
    const base = this.visit(node.children[0], value);
    if (!Array.isArray(base)) {
        return null;
    }
    const comparatorNode = node.children[2];
    const collected = [];
    for (const item of base) {
        if ((0, utils_js_1.isTruthy)(this.visit(comparatorNode, item))) {
            const current = this.visit(node.children[1], item);
            if (current !== null) {
                collected.push(current);
            }
        }
    }
    return collected;
}, _TreeInterpreter_visitFlatten = function _TreeInterpreter_visitFlatten(node, value) {
    const base = this.visit(node.children[0], value);
    if (!Array.isArray(base)) {
        return null;
    }
    const mergedList = [];
    for (const item of base) {
        if (Array.isArray(item)) {
            mergedList.push(...item);
        }
        else {
            mergedList.push(item);
        }
    }
    return mergedList;
}, _TreeInterpreter_visitIdentity = function _TreeInterpreter_visitIdentity(_node, value) {
    return value;
}, _TreeInterpreter_visitIndex = function _TreeInterpreter_visitIndex(node, value) {
    if (!Array.isArray(value)) {
        return null;
    }
    // The Python implementation doesn't support string indexing
    // even though we could, so we won't either for now.
    if (typeof node.value !== 'number') {
        throw new errors_js_1.JMESPathError(`Invalid index: ${node.value}`);
    }
    const index = node.value < 0 ? value.length + node.value : node.value;
    const found = value[index];
    if (found === undefined) {
        return null;
    }
    return found;
}, _TreeInterpreter_visitSlice = function _TreeInterpreter_visitSlice(node, value) {
    const step = (0, typeutils_1.isIntegerNumber)(node.children[2]) ? node.children[2] : 1;
    if (step === 0) {
        throw new Error('Invalid slice, step cannot be 0');
    }
    if (!Array.isArray(value)) {
        return null;
    }
    if (value.length === 0) {
        return [];
    }
    return (0, utils_js_1.sliceArray)({
        array: value,
        start: node.children[0],
        end: node.children[1],
        step,
    });
}, _TreeInterpreter_visitKeyValPair = function _TreeInterpreter_visitKeyValPair(node, value) {
    return this.visit(node.children[0], value);
}, _TreeInterpreter_visitLiteral = function _TreeInterpreter_visitLiteral(node, _value) {
    return node.value;
}, _TreeInterpreter_visitMultiSelectObject = function _TreeInterpreter_visitMultiSelectObject(node, value) {
    if (Object.is(value, null)) {
        return null;
    }
    const collected = {};
    for (const child of node.children) {
        if (typeof child.value === 'string') {
            collected[child.value] = this.visit(child, value);
        }
    }
    return collected;
}, _TreeInterpreter_visitMultiSelectList = function _TreeInterpreter_visitMultiSelectList(node, value) {
    if (Object.is(value, null)) {
        return null;
    }
    const collected = [];
    for (const child of node.children) {
        collected.push(this.visit(child, value));
    }
    return collected;
}, _TreeInterpreter_visitOrExpression = function _TreeInterpreter_visitOrExpression(node, value) {
    const matched = this.visit(node.children[0], value);
    if (!(0, utils_js_1.isTruthy)(matched)) {
        return this.visit(node.children[1], value);
    }
    return matched;
}, _TreeInterpreter_visitAndExpression = function _TreeInterpreter_visitAndExpression(node, value) {
    const matched = this.visit(node.children[0], value);
    if (!(0, utils_js_1.isTruthy)(matched)) {
        return matched;
    }
    return this.visit(node.children[1], value);
}, _TreeInterpreter_visitNotExpression = function _TreeInterpreter_visitNotExpression(node, value) {
    const originalResult = this.visit(node.children[0], value);
    if (typeof originalResult === 'number' && originalResult === 0) {
        // Special case for 0, !0 should be false, not true.
        // 0 is not a special cased integer in jmespath.
        return false;
    }
    return !(0, utils_js_1.isTruthy)(originalResult);
}, _TreeInterpreter_visitProjection = function _TreeInterpreter_visitProjection(node, value) {
    const base = this.visit(node.children[0], value);
    if (!Array.isArray(base)) {
        return null;
    }
    const collected = [];
    for (const item of base) {
        const current = this.visit(node.children[1], item);
        if (current !== null) {
            collected.push(current);
        }
    }
    return collected;
}, _TreeInterpreter_visitValueProjection = function _TreeInterpreter_visitValueProjection(node, value) {
    const base = this.visit(node.children[0], value);
    if (!(0, typeutils_1.isRecord)(base)) {
        return null;
    }
    const values = Object.values(base);
    const collected = [];
    for (const item of values) {
        const current = this.visit(node.children[1], item);
        if (current !== null) {
            collected.push(current);
        }
    }
    return collected;
};
