import { TreeInterpreter } from './TreeInterpreter.js';
import { ArityError, JMESPathTypeError, UnknownFunctionError, VariadicArityError, } from './errors.js';
class ParsedResult {
    constructor(expression, parsed) {
        this.expression = expression;
        this.parsed = parsed;
    }
    /**
     * Perform a JMESPath search on a JSON value.
     *
     * @param value The JSON value to search
     * @param options The parsing options to use
     */
    search(value, options) {
        const interpreter = new TreeInterpreter(options);
        try {
            return interpreter.visit(this.parsed, value);
        }
        catch (error) {
            if (error instanceof JMESPathTypeError ||
                error instanceof UnknownFunctionError ||
                error instanceof ArityError ||
                error instanceof VariadicArityError) {
                error.setExpression(this.expression);
            }
            throw error;
        }
    }
}
export { ParsedResult };
